// ABBYY® Mobile Capture © 2019 ABBYY Production LLC.
// ABBYY is a registered trademark or a trademark of ABBYY Software Ltd.

import UIKit
import AbbyyRtrSDK

class DataLoadOperation: Operation {
	var image: UIImage?
	var loadingCompleteHandler: ((UIImage?) -> Void)?

	fileprivate let url: URL

	init(imageUrl: URL)
	{
		url = imageUrl
	}

	override func main()
	{
		if isCancelled { return }

		if let data = try? Data(contentsOf: url) {
			image = UIImage(data: data)
		}

		if isCancelled { return }

		DispatchQueue.main.async { [weak self] in
			self?.loadingCompleteHandler?(self?.image)
		}
	}
}

//MARK: Utility class for working with image export using Mobile Capture SDK
class ImageExporter {
	enum ExportError: Error {
		case invalidLicense
		case writingFailure
		case cleanFailure
		case internalError
	}

	struct PdfExportStatus {
		enum Action: String {
			case fetching
			case writing
		}
		let pagesProcessed: Int
		let pagesCount: Int
		let currentAction: Action
	}

	var tmpDirectory = FileManager.default.temporaryDirectory

	func exportPdf(
		filename: String,
		pages: [UIImage],
		progress: ((PdfExportStatus) -> ())?) throws
	{
		guard let coreApi = RecognitionEngine.shared?.createCoreAPI() else {
			throw ExportError.invalidLicense
		}
		do {
			let cache = try FileManager.default.contentsOfDirectory(atPath: tmpDirectory.path)
			for filename in cache {
				let filepath = tmpDirectory.appendingPathComponent(filename)
				try FileManager.default.removeItem(atPath: filepath.path)
			}
		} catch {
			throw ExportError.cleanFailure
		}
		let tmpUrl = tmpDirectory.appendingPathComponent(filename)
		let stream = RTRFileOutputStream(filePath: tmpUrl.path)
		if let error = stream.error {
			throw error
		}
		let exportOperation = coreApi.createExport(toPdfOperation: stream)
		for (page, pageIndex) in zip(pages, 0 ..< pages.count) {
			let status = PdfExportStatus(
				pagesProcessed: pageIndex,
				pagesCount: pages.count,
				currentAction: .writing)
			progress?(status)
			guard exportOperation.addPage(with: page) else {
				throw ExportError.writingFailure
			}
		}
		guard exportOperation.close() else {
			throw ExportError.writingFailure
		}
	}

	func exportPdf(
		filename: String,
		pages: [URL],
		progress: ((PdfExportStatus) -> ())?) throws
	{
		let images = zip(pages, 0 ..< pages.count).compactMap { (url, index) -> UIImage? in
			if let data = try? Data(contentsOf: url) {
				let exportStatus = PdfExportStatus(
					pagesProcessed: index,
					pagesCount: pages.count,
					currentAction: .fetching)
				progress?(exportStatus)
				return UIImage(data: data)
			}
			return .none
		}
		try exportPdf(filename: filename, pages: images, progress: progress)
	}

	func exportImage(destination: URL, page: UIImage) throws
	{
		guard let coreApi = RecognitionEngine.shared?.createCoreAPI() else {
			throw ExportError.invalidLicense
		}
		let stream = RTRFileOutputStream(filePath: destination.path)
		if let error = stream.error {
			throw error
		}
		let exportOperation = coreApi.createExport(toJpgOperation: stream)
		exportOperation.compression = .normalLevel

		guard exportOperation.addPage(with: page), exportOperation.close() else {
			throw ExportError.writingFailure
		}
	}
}

//MARK: Utility class for working with captured documents
class DataStorage {
	let loadingQueue = OperationQueue()
	let exporter = ImageExporter()
	var loadingOperations: [Int: DataLoadOperation] = [:]

	fileprivate lazy var directory: URL = {
		return FileManager.default.urls(for: .documentDirectory, in: .userDomainMask)[0]
	}()

	private(set) lazy var content: [URL] = {
		let fileManager = FileManager.default
		let unorderedContent = (try? fileManager.contentsOfDirectory(atPath: directory.path)) ?? []

		let creationDate: (URL) -> (Date) = {
			let date = (try? fileManager.attributesOfItem(atPath: $0.path)[.creationDate]) as? Date
			return date ?? Date()
		}
		return unorderedContent.map{ directory.appendingPathComponent($0) }.sorted{ creationDate($0) < creationDate($1) }
	}()

	var itemsCount: Int {
		return content.count
	}

	func clear()
	{
		for operation in loadingOperations.values {
			operation.cancel()
		}
		loadingOperations.removeAll()
		try? FileManager.default.removeItem(at: directory)
		content = []
	}

	func save(image: UIImage) throws
	{
		let filename = "\(String(describing: Date().timeIntervalSince1970)).jpg"
		let urlToSave = directory.appendingPathComponent(filename)
		try exporter.exportImage(destination: urlToSave, page: image)
		content.append(urlToSave)
	}

	func exportAsPdf(
		filename: String,
		progress: ((ImageExporter.PdfExportStatus) -> ())?) throws
	{
		try exporter.exportPdf(
			filename: filename,
			pages: content,
			progress: progress)
	}

	func fetch(at index: Int) -> DataLoadOperation?
	{
		if (0 ..< content.count).contains(index) {
			return DataLoadOperation(imageUrl: content[index])
		}
		return .none
	}

	func removeLast(count: Int = 1)
	{
		for _ in 0 ..< min(count, content.count) {
			if let url = content.last {
				do {
					try FileManager.default.removeItem(at: url)
					content.removeLast()
				} catch {
					continue
				}
			}
		}
	}
}
